import { validateAndParseAddress } from './utils';
import { TokenAmount } from './entities/fractions/tokenAmount'
import { Price } from './entities/fractions/price';
import { getSignatureWithProvider, getSignature, Message, getTypedData } from './entities/eip712'
import { ChainId } from './constants';
import { ethers } from 'ethers';
export class LimitOrder {

    public readonly maker: string;
    public readonly amountIn: TokenAmount;
    public readonly amountOut: TokenAmount;
    public readonly recipient: string;
    public readonly startTime: string;
    public readonly endTime: string;
    public readonly stopPrice: string;
    public readonly oracleAddress: string;
    public readonly oracleData: string;
    public v: number;
    public r: string;
    public s: string;
    
    constructor(
        maker: string,
        amountIn: TokenAmount,
        amountOut: TokenAmount,
        recipient: string,
        startTime: string,
        endTime: string,
        stopPrice: string,
        oracleAddress: string,
        oracleData: string
    ) { 
        this.maker = validateAndParseAddress(maker);
        this.amountIn = amountIn;
        this.amountOut = amountOut;
        this.recipient = validateAndParseAddress(recipient);
        this.startTime = startTime;
        this.endTime = endTime;
        this.stopPrice = stopPrice;
        this.oracleAddress = validateAndParseAddress(oracleAddress);
        this.oracleData = oracleData;
        this.v = 0;
        this.r = "";
        this.s = "";
    }

    // use price of token 
    usePrice(price: Price): LimitOrder {
        return new LimitOrder(
            this.maker,
            this.amountIn,
            new TokenAmount(this.amountOut.token, price.quote(this.amountIn).raw.toString()),
            this.recipient,
            this.startTime,
            this.endTime,
            this.stopPrice,
            this.oracleAddress,
            this.oracleData
        );
    }

    signdOrderWithPrivatekey(chainId: ChainId, privateKey: string) {
        let order : Message = {
            maker: this.maker,
            tokenIn: this.amountIn.token.address,
            tokenOut: this.amountOut.token.address,
            amountIn: this.amountIn.raw.toString(),
            amountOut: this.amountOut.raw.toString(),
            recipient: this.recipient,
            startTime: this.startTime,
            endTime: this.endTime,
            stopPrice: this.stopPrice,
            oracleAddress : this.oracleAddress,
            oracleData : this.oracleData
        }

        const { v, r, s } = getSignature(order, chainId, privateKey);

        this.v = v;
        this.r = r;
        this.s = s;

        return { v, r, s }
    }

    async signOrderWithProvider(chainId: ChainId, provider: ethers.providers.Web3Provider) {
        let order: Message = {
            maker: this.maker,
            tokenIn: this.amountIn.token.address,
            tokenOut: this.amountOut.token.address,
            amountIn: this.amountIn.raw.toString(),
            amountOut: this.amountOut.raw.toString(),
            recipient: this.recipient,
            startTime: this.startTime,
            endTime: this.endTime,
            stopPrice: this.stopPrice,
            oracleAddress: this.oracleAddress,
            oracleData: this.oracleData
        }

        const { v, r, s } = await getSignatureWithProvider(order, chainId, provider);

        this.v = v;
        this.r = r;
        this.s = s;

        return { v, r, s }
    }

    getTypedData(chainId: ChainId) {

        let order : Message = {
            maker: this.maker,
            tokenIn: this.amountIn.token.address,
            tokenOut: this.amountOut.token.address,
            amountIn: this.amountIn.raw.toString(),
            amountOut: this.amountOut.raw.toString(),
            recipient : this.recipient,
            startTime: this.startTime,
            endTime: this.endTime,
            stopPrice: this.stopPrice,
            oracleAddress : this.oracleAddress,
            oracleData : this.oracleData
        }

        return getTypedData(order, chainId);
    }
}
